#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>


#define DEFAULTSIZE 16

void fillRandomMatrix( double**, int );
void displayMatrix( double**, int );
void resetMatrix( double**, int );
double** allocMatrix( int );
double** freeMatrix( double**, int );
double getTime( void );
void printResult( char*, int, double, int );

void naiveMatMul( double**, double**, double**, int );
void tiledMatMul( double**, double**, double**, int, int );

void swappedMatMul( double**, double**, double**, int );
void tiledSwappedMatMul( double**, double**, double**, int, int );
void tiledSwappedMatMul2( double**, double**, double**, int, int );

int main( int argc, char** argv ) {

    int size;
    int block, curr_b;
    double start, end;
    double flops;

    double** A;
    double** B;
    double** C;

    if( argc > 1 ) {
        size = atoi( argv[1] );
        if( argc > 2 ) {
            block = atoi( argv[2] );
        } else {
            block = size;
        }
    } else {
        size = DEFAULTSIZE;
        block = size;
    }

    printf ( "size: %d\n", size );

    srand48( getpid() );

    A = allocMatrix( size );
    B = allocMatrix( size );
    C = allocMatrix( size );

    fillRandomMatrix( A, size );
    fillRandomMatrix( B, size );
    resetMatrix( C, size );

    if( size < 8 ) {
        displayMatrix( A, size );
        displayMatrix( B, size );
    }
    
    flops = size*size*size;

    start = getTime();
    naiveMatMul( A, B, C, size );
    end = getTime();
    
    printResult( "NAIVE", size, end-start, -1 );
    if( size < 8 ) {
        displayMatrix( C, size );
    }
    resetMatrix( C, size );
    
    start = getTime();
    swappedMatMul( A, B, C, size );
    end = getTime();
    
    printResult( "SWAPPED", size, end-start, -1 );
    for( curr_b = 1 ; curr_b <= block ; curr_b<<=1 ) {
        resetMatrix( C, size );
        
        start = getTime();
        tiledMatMul( A, B, C, size, curr_b );
        end = getTime();
        
        printResult( "TILED", size, end-start, curr_b );
        if( size < 8 ) {
            displayMatrix( C, size );
        }
    }
    
    for( curr_b = 16 ; curr_b <= block ; curr_b<<=1 ) {
        resetMatrix( C, size );
        
        start = getTime();
        tiledSwappedMatMul( A, B, C, size, curr_b );
        end = getTime();
        
        printResult( "TILED SWAPPED", size, end-start, curr_b );
        if( size < 8 ) {
            displayMatrix( C, size );
        }
    }
    
    for( curr_b = 16; curr_b <= block ; curr_b<<=1 ) {        
        resetMatrix( C, size );
        
        start = getTime();
        tiledSwappedMatMul2( A, B, C, size, curr_b );
        end = getTime();
        
        printResult( "TILED SWAPPED2", size, end-start, curr_b );
        if( size < 8 ) {
            displayMatrix( C, size );
        }
    }
    
    if( size < 8 ) {
        displayMatrix( C, size );
    }

    A = freeMatrix( A, size );
    B = freeMatrix( B, size );
    C = freeMatrix( C, size );

    return EXIT_SUCCESS;        

}

void printResult( char* algo, int size, double time, int tile ) {
    double flops = size*size*size;
    printf( "%-20s", algo );
    printf( "%d", size );
    if( -1 == tile ) {
        printf( "%-5s", "  " );
    } else {
        printf( "%*d", 5, tile );
    }
    printf( "\t   | Time: %12.0lf ns - flops: %12.0lf - Mflops/s: %8.2lf\n", time, flops, flops/(time/1000.) ) ;
}

void fillRandomMatrix( double** mat, int size ) {
    int i, j;
    for( i = 0 ; i < size ; i++ ) {
        for( j = 0 ; j < size ; j++ ) {
            mat[i][j] = drand48();
        }
    }    
}

void displayMatrix( double** mat, int size ) {
    int i, j;
    for( i = 0 ; i < size ; i++ ) {
        for( j = 0 ; j < size ; j++ ) {
            printf( "%.4lf   ", mat[i][j] );
        }
        printf( "\n" );
    }    
        printf( "\n" );
}

double** allocMatrix( int size ) {
    double** matrix;
    int i;
    matrix = (double**) malloc( size * sizeof( double* ) );
    for( i = 0 ; i < size ; i++ ) {
        matrix[i] = (double*) malloc( size * sizeof( double ) );
    }
    return matrix;
}

double** freeMatrix( double** mat, int size ) {
    int i;
    for( i = 0 ; i < size ; i++ ) {
        free( mat[i] );
        mat[i] = NULL;
    }
    free( mat );
    mat = NULL;
    return mat;
}

double getTime( ) {
    struct timespec ts;
    clock_gettime( CLOCK_REALTIME, &ts );
    return ( ts.tv_nsec + ts.tv_sec * 1000000000. );
}

void resetMatrix( double** mat, int size ) {
    int i;
    for( i = 0 ; i < size ; i++ ) {
        memset( (void*) mat[i], 0, size*sizeof( double ) );
    }
}


void naiveMatMul( double** A, double** B, double** C, int size ) {
    int i, j, k;
    for( i = 0 ; i < size ; i++ ) {
        for( j = 0 ; j < size ; j++ ) {
            for( k = 0 ; k < size ; k++ ) {
                C[i][j] += A[i][k] * B[k][j];
            }
        }
    }
}

void swappedMatMul( double** A, double** B, double** C, int size ) {
    int i, j, k;
    for( i = 0 ; i < size ; i++ ) {
        for( k = 0 ; k < size ; k++ ) {
            for( j = 0 ; j < size ; j++ ) {
                C[i][j] += A[i][k] * B[k][j];
            }
        }
    }
}


void tiledMatMul( double** A, double** B, double** C, int size, int block ) {
    int i, j, k, b, c;
    int theblock1, theblock2;
    for( i = 0 ; i < size ; i+=block ) {
        for( j = 0 ; j < size ; j+=block ) {

            theblock1 = ( j+block < size ) ? block : (size-j) ;
            theblock2 = ( i+block < size ) ? block : (size-i) ;
                        
            for( b = 0 ; b < theblock2 ; b++ ) {
                for( c = 0 ; c < theblock1 ; c++ ) {
                    
                    for( k = 0 ; k < size ; k++ ) {
                        C[i+b][j+c] += A[i+b][k] * B[k][j+c];
                    }
                }
            }
        }
    }
}

void tiledSwappedMatMul( double** A, double** B, double** C, int size, int block ) {
    int i, j, k, b, c;
    int theblock1, theblock2;
    for( i = 0 ; i < size ; i+=block ) {
        for( k = 0 ; k < size ; k++ ) {
            for( j = 0 ; j < size ; j+=block ) {

                theblock1 = ( j+block < size ) ? block : (size-j) ;
                theblock2 = ( i+block < size ) ? block : (size-i) ;

                for( b = 0 ; b < theblock2 ; b++ ) {
                    for( c = 0 ; c < theblock1 ; c++ ) {
                        C[i+b][j+c] += A[i+b][k] * B[k][j+c];
                    }
                }
            }
        }
    }
}

void tiledSwappedMatMul2( double** A, double** B, double** C, int size, int block ) {
    int i, j, k, b, c;
    int theblock1, theblock2;
    for( i = 0 ; i < size ; i+=block ) {
        for( j = 0 ; j < size ; j+=block ) {

            theblock1 = ( j+block < size ) ? block : (size-j) ;
            theblock2 = ( i+block < size ) ? block : (size-i) ;

            for( b = 0 ; b < theblock2 ; b++ ) {
                for( k = 0 ; k < size ; k++ ) {
                    for( c = 0 ; c < theblock1 ; c++ ) {
                        C[i+b][j+c] += A[i+b][k] * B[k][j+c];
                    }
                }
            }
        }
    }
}
