#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <papi.h>

#define DEFAULTSIZE 16
#define PAPICNT  3

void fillRandomMatrix( double**, int );
void displayMatrix( double**, int );
void resetMatrix( double**, int );
double** allocMatrix( int );
double** freeMatrix( double**, int );
double getTime( void );
void printResult( char*, int, double, int, long long int* );

void naiveMatMul( double**, double**, double**, int );
void tiledMatMul( double**, double**, double**, int, int );

void swappedMatMul( double**, double**, double**, int );
void tiledSwappedMatMul( double**, double**, double**, int, int );
void tiledSwappedMatMul2( double**, double**, double**, int, int );

void errorPAPI( int );
int initPAPI( int* );

#define STARTPAPI do {   \
        start = getTime();                            \
        rc = PAPI_read_counters( values, PAPICNT );   \
        if( rc != PAPI_OK ) errorPAPI( rc );          \
    } while( 0 );

#define STOPPAPI do {    \
        rc = PAPI_read_counters( values, PAPICNT );   \
        if( rc != PAPI_OK ) errorPAPI( rc );          \
        end = getTime();                              \
        printResult( algo, size, end-start, curr_b, values ); \
    } while( 0 );

int main( int argc, char** argv ) {

    int size;
    int block, curr_b;
    double start, end;
    double flops;
    char* algo;

    double** A;
    double** B;
    double** C;

    long long int values[PAPICNT];
    int events[PAPICNT];
    int rc;

    if( argc > 1 ) {
        size = atoi( argv[1] );
        if( argc > 2 ) {
            block = atoi( argv[2] );
        } else {
            block = size;
        }
    } else {
        size = DEFAULTSIZE;
        block = size;
    }

    printf ( "size: %d\n", size );
    rc = initPAPI( events );
    if( 0 > rc ) {
        printf( "Error during PAPI initialization\n" );
        return EXIT_FAILURE;
    }
    rc = PAPI_start_counters( events, PAPICNT );
        
    srand48( getpid() );

    A = allocMatrix( size );
    B = allocMatrix( size );
    C = allocMatrix( size );

    fillRandomMatrix( A, size );
    fillRandomMatrix( B, size );
    resetMatrix( C, size );

    if( size < 8 ) {
        displayMatrix( A, size );
        displayMatrix( B, size );
    }
    
    flops = size*size*size;
    curr_b = -1;

    algo = "NAIVE";
    STARTPAPI;
    naiveMatMul( A, B, C, size );
    STOPPAPI;
    
    if( size < 8 ) {
        displayMatrix( C, size );
    }
    resetMatrix( C, size );
    
    algo = "SWAPPED";
    STARTPAPI;
    swappedMatMul( A, B, C, size );
    STOPPAPI;
    
    for( curr_b = 1 ; curr_b <= block ; curr_b<<=1 ) {
        resetMatrix( C, size );
        
        algo = "TILED";
        STARTPAPI;
        tiledMatMul( A, B, C, size, curr_b );
        STOPPAPI;
        
        if( size < 8 ) {
            displayMatrix( C, size );
        }
    }
    
    for( curr_b = 16 ; curr_b <= block ; curr_b<<=1 ) {
        resetMatrix( C, size );
        
        algo = "TILED SWAPPED";
        STARTPAPI;
        tiledSwappedMatMul( A, B, C, size, curr_b );
        STOPPAPI;
        
        if( size < 8 ) {
            displayMatrix( C, size );
        }
    }
    
    for( curr_b = 16; curr_b <= block ; curr_b<<=1 ) {        
        resetMatrix( C, size );

        algo = "TILED SWAPPED2";
        STARTPAPI;
        tiledSwappedMatMul2( A, B, C, size, curr_b );
        STOPPAPI;
        
        if( size < 8 ) {
            displayMatrix( C, size );
        }
    }
    
    A = freeMatrix( A, size );
    B = freeMatrix( B, size );
    C = freeMatrix( C, size );

    return EXIT_SUCCESS;        

}

void printResult( char* algo, int size, double time, int tile, long long int* values ) {
    int i;
    double flops = size*size*size;
    printf( "%-20s", algo );
    printf( "%d", size );
    if( -1 == tile ) {
        printf( "%-5s", "  " );
    } else {
        printf( "%*d", 5, tile );
    }
    printf( "\t   | Time: %12.0lf ns - flops: %12.0lf - Mflops/s: %8.2lf", time, flops, flops/(time/1000.) ) ;
    printf( " | PAPI " );
    for( i = 0 ; i < PAPICNT ; i++ ) {
        printf( "%12lld  ", values[i] );
    }
    printf( "\n" );
}

void fillRandomMatrix( double** mat, int size ) {
    int i, j;
    for( i = 0 ; i < size ; i++ ) {
        for( j = 0 ; j < size ; j++ ) {
            mat[i][j] = drand48();
        }
    }    
}

void displayMatrix( double** mat, int size ) {
    int i, j;
    for( i = 0 ; i < size ; i++ ) {
        for( j = 0 ; j < size ; j++ ) {
            printf( "%.4lf   ", mat[i][j] );
        }
        printf( "\n" );
    }    
        printf( "\n" );
}

double** allocMatrix( int size ) {
    double** matrix;
    int i;
    matrix = (double**) malloc( size * sizeof( double* ) );
    for( i = 0 ; i < size ; i++ ) {
        matrix[i] = (double*) malloc( size * sizeof( double ) );
    }
    return matrix;
}

double** freeMatrix( double** mat, int size ) {
    int i;
    for( i = 0 ; i < size ; i++ ) {
        free( mat[i] );
        mat[i] = NULL;
    }
    free( mat );
    mat = NULL;
    return mat;
}

double getTime( ) {
    struct timespec ts;
    clock_gettime( CLOCK_REALTIME, &ts );
    return ( ts.tv_nsec + ts.tv_sec * 1000000000. );
}

void resetMatrix( double** mat, int size ) {
    int i;
    for( i = 0 ; i < size ; i++ ) {
        memset( (void*) mat[i], 0, size*sizeof( double ) );
    }
}

void naiveMatMul( double** A, double** B, double** C, int size ) {
    int i, j, k;
    for( i = 0 ; i < size ; i++ ) {
        for( j = 0 ; j < size ; j++ ) {
            for( k = 0 ; k < size ; k++ ) {
                C[i][j] += A[i][k] * B[k][j];
            }
        }
    }
}

void swappedMatMul( double** A, double** B, double** C, int size ) {
    int i, j, k;
    for( i = 0 ; i < size ; i++ ) {
        for( k = 0 ; k < size ; k++ ) {
            for( j = 0 ; j < size ; j++ ) {
                C[i][j] += A[i][k] * B[k][j];
            }
        }
    }
}


void tiledMatMul( double** A, double** B, double** C, int size, int block ) {
    int i, j, k, b, c;
    int theblock1, theblock2;
    for( i = 0 ; i < size ; i+=block ) {
        for( j = 0 ; j < size ; j+=block ) {

            theblock1 = ( j+block < size ) ? block : (size-j) ;
            theblock2 = ( i+block < size ) ? block : (size-i) ;
                        
            for( b = 0 ; b < theblock2 ; b++ ) {
                for( c = 0 ; c < theblock1 ; c++ ) {
                    
                    for( k = 0 ; k < size ; k++ ) {
                        C[i+b][j+c] += A[i+b][k] * B[k][j+c];
                    }
                }
            }
        }
    }
}

void tiledSwappedMatMul( double** A, double** B, double** C, int size, int block ) {
    int i, j, k, b, c;
    int theblock1, theblock2;
    for( i = 0 ; i < size ; i+=block ) {
        for( k = 0 ; k < size ; k++ ) {
            for( j = 0 ; j < size ; j+=block ) {

                theblock1 = ( j+block < size ) ? block : (size-j) ;
                theblock2 = ( i+block < size ) ? block : (size-i) ;

                for( b = 0 ; b < theblock2 ; b++ ) {
                    for( c = 0 ; c < theblock1 ; c++ ) {
                        C[i+b][j+c] += A[i+b][k] * B[k][j+c];
                    }
                }
            }
        }
    }
}

void tiledSwappedMatMul2( double** A, double** B, double** C, int size, int block ) {
    int i, j, k, b, c;
    int theblock1, theblock2;
    for( i = 0 ; i < size ; i+=block ) {
        for( j = 0 ; j < size ; j+=block ) {

            theblock1 = ( j+block < size ) ? block : (size-j) ;
            theblock2 = ( i+block < size ) ? block : (size-i) ;

            for( b = 0 ; b < theblock2 ; b++ ) {
                for( k = 0 ; k < size ; k++ ) {
                    for( c = 0 ; c < theblock1 ; c++ ) {
                        C[i+b][j+c] += A[i+b][k] * B[k][j+c];
                    }
                }
            }
        }
    }
}

void errorPAPI( int error ){

    printf( "error starting counting events\n" );
    switch( error ){
    case PAPI_EINVAL:
      printf( "One or more of the arguments is invalid.\n" );
      break;
    case PAPI_EISRUN:
      printf( "Counters have already been started, you must call PAPI_stop_counters() before you call this function again.\n" );
      break;
    case PAPI_ESYS:
      printf( "A system or C library call failed inside PAPI, see the errno variable.\n" );
      break;
    case PAPI_ENOMEM:
      printf( "Insufficient memory to complete the operation.\n" );
      break;
    case PAPI_ECNFLCT:
      printf( "The underlying counter hardware cannot count this event and other events in the EventSet simultaneously.\n" );
    case PAPI_ENOEVNT:
      printf( "The PAPI preset is not available on the underlying hardware.\n" );
      break;
    }

}

int initPAPI( int* events ){
    int num_hwcntrs;
    int rc;
    
    events[0] = PAPI_L1_DCM;
    events[1] = PAPI_L2_DCM;
    events[2] = PAPI_L3_TCM; /* DCM pas dispo */

    if( ( num_hwcntrs = PAPI_num_counters()) <= PAPI_OK
        || num_hwcntrs < PAPICNT ){
        printf( "Error while retrieving counters info\n" );
        return -1;
    } 
    return 0;
}
